(fluid) {

	//seed points

	(float) {
		id = "groups.count";
		value = 200;
	}

	(float) {
		id = "groupsize.count";
		value = 13;
	}

	(object) {
		id = "positions1.obj";
		mesh = "positions1.mesh";
		(passes) {
			"pos.advect.pass"
			"vel.pass"
		}
	}

	(object) {
		id = "positions2.obj";
		mesh = "positions2.mesh";
		(passes) {
			"pos.copyback.pass"
		}
	}

	(pass) {
		id = "pos.advect.pass";
		shader = "pos.advect.shader";
		(floats) {
			D_time = "D_time.float";
		}
		(textures) {
			velocity = "fluid.velocitypressure.texture";
		}
		geometryoutput = "positions2.mesh";
	}

	(pass) {
		id = "pos.copyback.pass";
		shader = "pos.copyback.shader";
		geometryoutput = "positions1.mesh";
	}

	(pass) {
		id = "vel.pass";
		shader = "vel.shader";
		(textures) {
			velocity = "fluid.velocitypressure.texture";
		}
		geometryoutput = "velocities.mesh";
	}

	(shader) {
		id = "pos.advect.shader";
		(transformfeedback) {
			"opos"
		}
		(vertex) {"
			#version 400 core
			
			in vec4 pos;
			out vec4 opos;

			uniform float D_time;
			uniform sampler2D velocity;
			
			void main() {
				//advect in time
				vec2 t_pos = (pos.xy + vec2(1.0))*0.5;
				vec2 vel = texture(velocity,t_pos).rg;
				t_pos = t_pos + vel*D_time;

				//determine step length for all the steps
				vec2 vel1 = texture(velocity,t_pos).rg;
				vec2 vel2 = texture(velocity,t_pos + vel*0.2*sign(pos.w)).rg;
				float l = length(vel1 + vel2) / 2.0;

				//advect in space (in order to form an arrow glyph)
				for(float i=abs(pos.w) ; i>0 ; i--) {
					vec2 vel1 = texture(velocity,t_pos).rg;
					vec2 vel2 = texture(velocity,t_pos + vel*0.2*sign(pos.w)).rg;
					vec2 vel = normalize(vec2(0.00011,0.00013)+vel1 + vel2)*l;
					t_pos = t_pos + vel*0.1*sign(pos.w);
				}

				//calculate vorticity (|nabla x (v.x,v.y,0)| )
				vec2 dx = vec2(0.01,0);
				vec2 dy = vec2(0,0.01);
				float vorticity = 
					texture2D(velocity,t_pos + dx).y - texture2D(velocity,t_pos - dx).y
					- (
					texture2D(velocity,t_pos + dy).x - texture2D(velocity,t_pos - dy).x
					);
				vorticity = clamp(abs(vorticity) * 100.0 , 0.0 , 1.0);

				opos = vec4(t_pos*2.0 - vec2(1.0) , vorticity , pos.w);
			}
			"
		}
	}

	(shader) {
		id = "pos.copyback.shader";
		(transformfeedback) {
			"opos"
		}
		(vertex) {"
			#version 400 core
			
			in vec4 pos;
			out vec4 opos;
			
			void main() {
				opos = pos;
			}
			"
		}
	}

	(shader) {
		id = "vel.shader";
		(transformfeedback) {
			"ovel"
		}
		(vertex) {"
			#version 400 core
			
			in vec4 pos;
			out vec4 ovel;

			uniform sampler2D velocity;
			
			void main() {
				vec2 t_pos = (pos.xy + vec2(1.0))*0.5;
				ovel = vec4(texture2D(velocity,t_pos).rg,0,1);
			}
			"
		}
	}

//DEBUG QUAD MESH

(mesh) {
	id = "quad.mesh";
	faceSize = 3;
	(attribute) {
		name = "arrow";
		outputName = "oarrow";
		attributeSize = 3;
		(data) {"
			-0.02 -0.02 -0.1
			0.02 0.02 -0.1
			-0.02 0.02 -0.1
			-0.02 -0.02 -0.1
			0.02 -0.02 -0.1
			0.02 0.02 -0.1
			"
		}
	}
}

(shader) {
	id = "quad.shader";
	(vertex) {"
		#version 400 core

		in vec3 arrow;

		in vec3 pos;

		void main() {
			gl_Position = vec4(arrow + pos,1);
		}
		"
	}
	(fragment) {"
		#version 400 core

		out vec4 fragment;

		void main() {
			fragment = vec4(1);
		}
		"
	}
}

(object) {
	id = "quad.obj";
	mesh = "quad.mesh";
	shader = "quad.shader";
	(instancebuffers) {
		"positions1.mesh"
	}
	(passes) {
//		"present.pass"
	}
}

	//glyph

	(matrix) {
		id = "obj2world.matrix";
		(value) {"
			0.02 0 0 0
			0 0.02 0 0
			0 0 0.02 -0.1
			0 0 0 1
			"
		}
	}

	(float) {
		id = "morph.float";
		value = 1;
	}

	//object containing all glyph instances

	(object) {
		id = "glyph.object";
		mesh = "glyph.mesh";
		shader = "glyph.shader";
		(matrices) {
			obj2world = "obj2world.matrix";
		}
		(floats) {
			groupsize = "groupsize.count";
		}
		(passes) {
			"presentarrow.pass"
		}
	}

	(shader) {
		id = "glyph.shader";
		(vertex) {"
			#version 400 core

			in vec3 arrow;
			in vec3 circle;
			in float weight;

			out float vorticity;

			uniform float groupsize;

			uniform float glyphsize;
			uniform vec4 sourcepoint1;
			uniform vec4 sourcepoint2;
			uniform vec4 sourcepoint3;
			uniform vec4 sourcepoint4;
			uniform vec4 sourcepoint5;
			uniform vec4 sourcepoint6;
			uniform vec4 sourcepoint7;
			uniform vec4 sourcepoint8;
			uniform vec4 sourcepoint9;
			uniform vec4 sourcepoint10;
			uniform vec4 sourcepoint11;
			uniform vec4 sourcepoint12;
			uniform vec4 sourcepoint13;
			uniform vec4 controlpoint1;
			uniform vec4 controlpoint2;
			uniform vec4 controlpoint3;
			uniform vec4 controlpoint4;
			uniform vec4 controlpoint5;
			uniform vec4 controlpoint6;
			uniform vec4 controlpoint7;
			uniform vec4 controlpoint8;
			uniform vec4 controlpoint9;
			uniform vec4 controlpoint10;
			uniform vec4 controlpoint11;
			uniform vec4 controlpoint12;
			uniform vec4 controlpoint13;
			uniform vec4 direction1;
			uniform vec4 direction2;
			uniform vec4 direction3;
			uniform vec4 direction4;
			uniform vec4 direction5;
			uniform vec4 direction6;
			uniform vec4 direction7;
			uniform vec4 direction8;
			uniform vec4 direction9;
			uniform vec4 direction10;
			uniform vec4 direction11;
			uniform vec4 direction12;
			uniform vec4 direction13;

			uniform mat4 obj2world;

			vec4 map_obj2world(vec4 obj_pos, vec4 source_pos, vec4 control_pos, vec4 direction) {
				float vort = control_pos.z;
				float vorticity_scaled = 1.0*(1.0-vort) + 0.4*vort;
				vec4 cpos = vec4(control_pos.x,control_pos.y,0,1);
				vec4 dir = vec4(normalize(direction.xyz),0);
				vec4 n_dir = vec4(-dir.y,dir.x,0,0);
				mat4 transform2world = mat4(dir, n_dir, 0,0,1,0, cpos);
				vec4 unchanged_obj = obj2world * (obj_pos - vec4(source_pos.xyz,0));
				unchanged_obj.xyz = unchanged_obj.xyz * vorticity_scaled;
				return transform2world * unchanged_obj;
			}

			float map_weight(float w, float pos, float n) {
				float lin_weight = 1 - clamp(abs(w-pos)*(n-1) , 0, 1);
				return cos((1-lin_weight)*3.14159)*0.5 + 0.5;
			}

			void main() {
				float n = groupsize;
				float w1 = map_weight(weight, 0, n);
				float w2 = map_weight(weight, 1/(n-1), n);
				float w3 = map_weight(weight, 2/(n-1), n);
				float w4 = map_weight(weight, 3/(n-1), n);
				float w5 = map_weight(weight, 4/(n-1), n);
				float w6 = map_weight(weight, 5/(n-1), n);
				float w7 = map_weight(weight, 6/(n-1), n);
				float w8 = map_weight(weight, 7/(n-1), n);
				float w9 = map_weight(weight, 8/(n-1), n);
				float w10 = map_weight(weight, 9/(n-1), n);
				float w11 = map_weight(weight, 10/(n-1), n);
				float w12 = map_weight(weight, 11/(n-1), n);
				float w13 = map_weight(weight, 12/(n-1), n);
				vec4 seg1 = map_obj2world(vec4(arrow,1), sourcepoint1, controlpoint1, direction1);
				vec4 seg2 = map_obj2world(vec4(arrow,1), sourcepoint2, controlpoint2, direction2);
				vec4 seg3 = map_obj2world(vec4(arrow,1), sourcepoint3, controlpoint3, direction3);
				vec4 seg4 = map_obj2world(vec4(arrow,1), sourcepoint4, controlpoint4, direction4);
				vec4 seg5 = map_obj2world(vec4(arrow,1), sourcepoint5, controlpoint5, direction5);
				vec4 seg6 = map_obj2world(vec4(arrow,1), sourcepoint6, controlpoint6, direction6);
				vec4 seg7 = map_obj2world(vec4(arrow,1), sourcepoint7, controlpoint7, direction7);
				vec4 seg8 = map_obj2world(vec4(arrow,1), sourcepoint8, controlpoint8, direction8);
				vec4 seg9 = map_obj2world(vec4(arrow,1), sourcepoint9, controlpoint9, direction9);
				vec4 seg10 = map_obj2world(vec4(arrow,1), sourcepoint10, controlpoint10, direction10);
				vec4 seg11 = map_obj2world(vec4(arrow,1), sourcepoint11, controlpoint11, direction11);
				vec4 seg12 = map_obj2world(vec4(arrow,1), sourcepoint12, controlpoint12, direction12);
				vec4 seg13 = map_obj2world(vec4(arrow,1), sourcepoint13, controlpoint13, direction13);
				vec4 circ = map_obj2world(vec4(circle,1), sourcepoint7, controlpoint7, direction7);

				vorticity = controlpoint7.z;

				float morph = clamp(glyphsize*25.0 - 1.0 , 0 , 1);
				gl_Position = morph * (seg1*w1 + seg2*w2 + seg3*w3 + seg4*w4 + seg5*w5 + seg6*w6 + seg7*w7 + seg8*w8 + seg9*w9 + seg10*w10 + seg11*w11 + seg12*w12 + seg13*w13) + (1.0-morph) * circ;
			}
			"
		}
		(fragment) {"
			#version 400 core

			out vec4 fragment;

			uniform float visibility;
			in float vorticity;

			void main() {
				if(visibility < 0.004) {
					discard;
				}
				fragment = vec4(vorticity,1-vorticity,0,visibility);
			}
			"
		}
	}

	//objects for drawing glyph shadows

	(object) {
		id = "glyphshadow.object";
		shader = "glyphshadow.shader";
		mesh = "glyph.mesh";
		(matrices) {
			obj2world = "obj2world.matrix";
		}
		(floats) {
			groupsize = "groupsize.count";
		}
		(passes) {
			"glyphshadow.pass"
		}
	}

	(shader) {
		id = "glyphshadow.shader";
		(vertex) {"
			#version 400 core

			in vec3 arrow;
			in vec3 circle;
			in float weight;

			uniform float groupsize;

			uniform float glyphsize;
			uniform vec4 sourcepoint1;
			uniform vec4 sourcepoint2;
			uniform vec4 sourcepoint3;
			uniform vec4 sourcepoint4;
			uniform vec4 sourcepoint5;
			uniform vec4 sourcepoint6;
			uniform vec4 sourcepoint7;
			uniform vec4 sourcepoint8;
			uniform vec4 sourcepoint9;
			uniform vec4 sourcepoint10;
			uniform vec4 sourcepoint11;
			uniform vec4 sourcepoint12;
			uniform vec4 sourcepoint13;
			uniform vec4 controlpoint1;
			uniform vec4 controlpoint2;
			uniform vec4 controlpoint3;
			uniform vec4 controlpoint4;
			uniform vec4 controlpoint5;
			uniform vec4 controlpoint6;
			uniform vec4 controlpoint7;
			uniform vec4 controlpoint8;
			uniform vec4 controlpoint9;
			uniform vec4 controlpoint10;
			uniform vec4 controlpoint11;
			uniform vec4 controlpoint12;
			uniform vec4 controlpoint13;
			uniform vec4 direction1;
			uniform vec4 direction2;
			uniform vec4 direction3;
			uniform vec4 direction4;
			uniform vec4 direction5;
			uniform vec4 direction6;
			uniform vec4 direction7;
			uniform vec4 direction8;
			uniform vec4 direction9;
			uniform vec4 direction10;
			uniform vec4 direction11;
			uniform vec4 direction12;
			uniform vec4 direction13;

			uniform mat4 obj2world;

			vec4 map_obj2world(vec4 obj_pos, vec4 source_pos, vec4 control_pos, vec4 direction) {
				float vort = control_pos.z;
				float vorticity_scaled = 1.0*(1.0-vort) + 0.4*vort;
				vec4 cpos = vec4(control_pos.x,control_pos.y,0,1);
				vec4 dir = vec4(normalize(direction.xyz),0);
				vec4 n_dir = vec4(-dir.y,dir.x,0,0);
				mat4 transform2world = mat4(dir, n_dir, 0,0,1,0, cpos);
				vec4 unchanged_obj = obj2world * (obj_pos - vec4(source_pos.xyz,0));
				unchanged_obj.xyz = unchanged_obj.xyz * vorticity_scaled;
				return transform2world * unchanged_obj;
			}

			float map_weight(float w, float pos, float n) {
				float lin_weight = 1 - clamp(abs(w-pos)*(n-1) , 0, 1);
				return cos((1-lin_weight)*3.14159)*0.5 + 0.5;
			}

			void main() {
				float n = groupsize;
				float w1 = map_weight(weight, 0, n);
				float w2 = map_weight(weight, 1/(n-1), n);
				float w3 = map_weight(weight, 2/(n-1), n);
				float w4 = map_weight(weight, 3/(n-1), n);
				float w5 = map_weight(weight, 4/(n-1), n);
				float w6 = map_weight(weight, 5/(n-1), n);
				float w7 = map_weight(weight, 6/(n-1), n);
				float w8 = map_weight(weight, 7/(n-1), n);
				float w9 = map_weight(weight, 8/(n-1), n);
				float w10 = map_weight(weight, 9/(n-1), n);
				float w11 = map_weight(weight, 10/(n-1), n);
				float w12 = map_weight(weight, 11/(n-1), n);
				float w13 = map_weight(weight, 12/(n-1), n);
				vec4 seg1 = map_obj2world(vec4(arrow,1), sourcepoint1, controlpoint1, direction1);
				vec4 seg2 = map_obj2world(vec4(arrow,1), sourcepoint2, controlpoint2, direction2);
				vec4 seg3 = map_obj2world(vec4(arrow,1), sourcepoint3, controlpoint3, direction3);
				vec4 seg4 = map_obj2world(vec4(arrow,1), sourcepoint4, controlpoint4, direction4);
				vec4 seg5 = map_obj2world(vec4(arrow,1), sourcepoint5, controlpoint5, direction5);
				vec4 seg6 = map_obj2world(vec4(arrow,1), sourcepoint6, controlpoint6, direction6);
				vec4 seg7 = map_obj2world(vec4(arrow,1), sourcepoint7, controlpoint7, direction7);
				vec4 seg8 = map_obj2world(vec4(arrow,1), sourcepoint8, controlpoint8, direction8);
				vec4 seg9 = map_obj2world(vec4(arrow,1), sourcepoint9, controlpoint9, direction9);
				vec4 seg10 = map_obj2world(vec4(arrow,1), sourcepoint10, controlpoint10, direction10);
				vec4 seg11 = map_obj2world(vec4(arrow,1), sourcepoint11, controlpoint11, direction11);
				vec4 seg12 = map_obj2world(vec4(arrow,1), sourcepoint12, controlpoint12, direction12);
				vec4 seg13 = map_obj2world(vec4(arrow,1), sourcepoint13, controlpoint13, direction13);
				vec4 circ = map_obj2world(vec4(circle,1), sourcepoint7, controlpoint7, direction7);

				float morph = clamp(glyphsize*25.0 - 1.0 , 0 , 1);
				gl_Position = morph * (seg1*w1 + seg2*w2 + seg3*w3 + seg4*w4 + seg5*w5 + seg6*w6 + seg7*w7 + seg8*w8 + seg9*w9 + seg10*w10 + seg11*w11 + seg12*w12 + seg13*w13) + (1.0-morph) * circ;
			}
			"
		}
		(fragment) {"
			#version 400 core

			out vec4 shadowmap;

			uniform float visibility;

			void main() {
				if(visibility < 0.004) {
					discard;
				}
				shadowmap = vec4(vec3(1),visibility);
			}
			"
		}
	}

	(pass) {
		id = "glyphshadow.pass";
		(output) {
			rendertarget = "present.offline.target";
			shadowmap = "glyphshadow.texture";
			targetBlendFactor = "one_minus_fragmentalpha";
			fragmentBlendFactor = "fragmentalpha";
		}
	}

	(shader) {
		id = "gauss.x.shader";
		(vertex) {"
			#version 400 core
			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords[15];

			uniform float kernelsize;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords[0] = texcoords + vec2(-1,0)*kernelsize;
				vtexcoords[1] = texcoords + vec2(-0.85714285,0)*kernelsize;
				vtexcoords[2] = texcoords + vec2(-0.71428571,0)*kernelsize;
				vtexcoords[3] = texcoords + vec2(-0.57142857,0)*kernelsize;
				vtexcoords[4] = texcoords + vec2(-0.42857142,0)*kernelsize;
				vtexcoords[5] = texcoords + vec2(-0.28571428,0)*kernelsize;
				vtexcoords[6] = texcoords + vec2(-0.14285714,0)*kernelsize;
				vtexcoords[7] = texcoords;
				vtexcoords[8] = texcoords + vec2(0.14285714,0)*kernelsize;
				vtexcoords[9] = texcoords + vec2(0.28571428,0)*kernelsize;
				vtexcoords[10] = texcoords + vec2(0.42857142,0)*kernelsize;
				vtexcoords[11] = texcoords + vec2(0.57142857,0)*kernelsize;
				vtexcoords[12] = texcoords + vec2(0.71428571,0)*kernelsize;
				vtexcoords[13] = texcoords + vec2(0.85714285,0)*kernelsize;
				vtexcoords[14] = texcoords + vec2(1,0)*kernelsize;
			}
			"
		}
		(fragment) {"
			#version 400 core
			in vec2 vtexcoords[15];
			out vec4 filtered;

			uniform sampler2D unfiltered;

			void main() {
				filtered += texture(unfiltered,vtexcoords[0])*0.00442991210551132;
				filtered += texture(unfiltered,vtexcoords[1])*0.00895781211794;
				filtered += texture(unfiltered,vtexcoords[2])*0.0215963866053;
				filtered += texture(unfiltered,vtexcoords[3])*0.0443683338718;
				filtered += texture(unfiltered,vtexcoords[4])*0.0776744219933;
				filtered += texture(unfiltered,vtexcoords[5])*0.115876621105;
				filtered += texture(unfiltered,vtexcoords[6])*0.147308056121;
				filtered += texture(unfiltered,vtexcoords[7])*0.159576912161;
				filtered += texture(unfiltered,vtexcoords[8])*0.147308056121;
				filtered += texture(unfiltered,vtexcoords[9])*0.115876621105;
				filtered += texture(unfiltered,vtexcoords[10])*0.0776744219933;
				filtered += texture(unfiltered,vtexcoords[11])*0.0443683338718;
				filtered += texture(unfiltered,vtexcoords[12])*0.0215963866053;
				filtered += texture(unfiltered,vtexcoords[13])*0.00895781211794;
				filtered += texture(unfiltered,vtexcoords[14])*0.00442991210551132;
			}
			"
		}
	}

	(shader) {
		id = "gauss.y.shader";
		(vertex) {"
			#version 400 core
			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords[15];

			uniform float kernelsize;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords[0] = texcoords + vec2(0,-1)*kernelsize;
				vtexcoords[1] = texcoords + vec2(0,-0.85714285)*kernelsize;
				vtexcoords[2] = texcoords + vec2(0,-0.71428571)*kernelsize;
				vtexcoords[3] = texcoords + vec2(0,-0.57142857)*kernelsize;
				vtexcoords[4] = texcoords + vec2(0,-0.42857142)*kernelsize;
				vtexcoords[5] = texcoords + vec2(0,-0.28571428)*kernelsize;
				vtexcoords[6] = texcoords + vec2(0,-0.14285714)*kernelsize;
				vtexcoords[7] = texcoords;
				vtexcoords[8] = texcoords + vec2(0,0.14285714)*kernelsize;
				vtexcoords[9] = texcoords + vec2(0,0.28571428)*kernelsize;
				vtexcoords[10] = texcoords + vec2(0,0.42857142)*kernelsize;
				vtexcoords[11] = texcoords + vec2(0,0.57142857)*kernelsize;
				vtexcoords[12] = texcoords + vec2(0,0.71428571)*kernelsize;
				vtexcoords[13] = texcoords + vec2(0,0.85714285)*kernelsize;
				vtexcoords[14] = texcoords + vec2(0,1)*kernelsize;
			}
			"
		}
		(fragment) {"
			#version 400 core
			in vec2 vtexcoords[15];
			out vec4 filtered;

			uniform sampler2D unfiltered;

			void main() {
				filtered += texture(unfiltered,vtexcoords[0])*0.00442991210551132;
				filtered += texture(unfiltered,vtexcoords[1])*0.00895781211794;
				filtered += texture(unfiltered,vtexcoords[2])*0.0215963866053;
				filtered += texture(unfiltered,vtexcoords[3])*0.0443683338718;
				filtered += texture(unfiltered,vtexcoords[4])*0.0776744219933;
				filtered += texture(unfiltered,vtexcoords[5])*0.115876621105;
				filtered += texture(unfiltered,vtexcoords[6])*0.147308056121;
				filtered += texture(unfiltered,vtexcoords[7])*0.159576912161;
				filtered += texture(unfiltered,vtexcoords[8])*0.147308056121;
				filtered += texture(unfiltered,vtexcoords[9])*0.115876621105;
				filtered += texture(unfiltered,vtexcoords[10])*0.0776744219933;
				filtered += texture(unfiltered,vtexcoords[11])*0.0443683338718;
				filtered += texture(unfiltered,vtexcoords[12])*0.0215963866053;
				filtered += texture(unfiltered,vtexcoords[13])*0.00895781211794;
				filtered += texture(unfiltered,vtexcoords[14])*0.00442991210551132;
			}
			"
		}
	}

	(float) {
		id = "gauss.kernelsize.float";
		value = 0.01;
	}

	(pass) {
		id = "gauss.x.pass";
		backgroundQuad = "true";
		shader = "gauss.x.shader";
		(floats) {
			kernelsize = "gauss.kernelsize.float";
		}
		(textures) {
			unfiltered = "glyphshadow.texture";
		}
		(output) {
			rendertarget = "present.offline.target";
			filtered = "glyphshadow2.texture";
		}
	}

	(pass) {
		id = "gauss.y.pass";
		backgroundQuad = "true";
		shader = "gauss.y.shader";
		(floats) {
			kernelsize = "gauss.kernelsize.float";
		}
		(textures) {
			unfiltered = "glyphshadow2.texture";
		}
		(output) {
			rendertarget = "present.offline.target";
			filtered = "glyphshadow.texture";
		}
	}

	//input

	(vector) {
		id = "displacement.vector";
	}

	(float) {
		id = "D_time.float";
	}

	(vector) {
		id = "control.velocity.vector";
	}

	(vector) {
		id = "control.position.vector";
	}

	(float) {
		id = "present.float";
		value = 1.0;
	}

	//fundamental data

	(vector) {
		id = "D_kernel.vector";
		(value) {"0.0015 0.0015"}
	}

	(texture) {
		id = "initfluid.texture";
		path = "../data/textures/water.png";
	}

	(texture) {
		id = "solid.texture";
		path = "../data/textures/quadworld.png";
	}

	(rendertarget) {
		id = "fluid.target";
		width = 1000;
		height = 1000;
		renderToDisplay = "false";
	}

	(texture) {
		id = "fluid.copy.texture";
		width = 1000;
		height = 1000;
		format = "float";
	}

	(texture) {
		id = "fluid.data.texture";
		width = 1000;
		height = 1000;
		format = "float";
	}

	(texture) {
		id = "fluid.velocitypressure.texture";
		width = 1000;
		height = 1000;
		format = "float";
	}

	(texture) {
		id = "fluid.velocitypressure2.texture";
		width = 1000;
		height = 1000;
		format = "float";
	}

	(texture) {
		id = "fluid.velocitypressure3.texture";
		width = 1000;
		height = 1000;
		format = "float";
	}

	//initialize

	(shader) {
		id = "copy.shader";
		(vertex) {"
			#version 400 core
			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
			}
			"
		}
		(fragment) {"
			#version 400 core
			in vec2 vtexcoords;
			out vec4 target;	

			uniform sampler2D source;

			void main() {
				target = texture2D(source,vtexcoords);
			}
			"
		}
	}

	(shader) {
		id = "black.shader";
		(vertex) {"
			#version 400 core
			in vec3 vertices;

			void main() {
				gl_Position = vec4(vertices,1);
			}
			"
		}
		(fragment) {"
			#version 400 core
			out vec4 target;

			void main() {
				target = vec4(0,0,0,0);
			}
			"
		}
	}

	(pass) {
		id = "init.data.pass";
		backgroundQuad = "true";
		shader = "copy.shader";
		(textures) {
			source = "initfluid.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			target = "fluid.data.texture";
		}
	}

	(pass) {
		id = "init.velocitypressure.pass";
		backgroundQuad = "true";
		shader = "black.shader";
		(output) {
			rendertarget = "fluid.target";
			target = "fluid.velocitypressure.texture";
		}
	}

	(effect) {
		id = "initfluid.effect";
		(passes) {
			"init.data.pass"
			"init.velocitypressure.pass"
		}
	}

	//calculate

	//change velocity field

	(shader) {
		id = "fluid.change.shader";
		(vertex) {"
			#version 400 core

			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;
			
			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
			}
			"
		}
		(fragment) {"
			#version 400 core

			in vec2 vtexcoords;
			out vec4 pressurevelocity_new;

			uniform float D_time;
			uniform vec4 gravity;
			uniform vec4 displacement;
			uniform sampler2D pressurevelocity_old;
			uniform sampler2D data;
			
			void main() {
				vec2 change = vec2(0.0);
				if(vtexcoords.x > 0.45 && vtexcoords.x < 0.55 && vtexcoords.y > 0.45 && vtexcoords.y < 0.55) {
					change = (vec2(0.1,0.1213)*0.0 + displacement.xy) * D_time;
				}
				change = change + texture2D(data,vtexcoords).b * vec2(0,-0.5) * D_time + texture2D(data,vtexcoords).r * vec2(0,0.1) * D_time;
				//float force = texture2D(data,vtexcoords).b;
				//change = change + (force * vec2(0,-0.4) + (1.0 - force) * vec2(0,0.4)) * D_time;
				pressurevelocity_new = texture2D(pressurevelocity_old,vtexcoords) + gravity * D_time;//vec4(change,0,0);
			}
			"
		}
	}

	(pass) {
		id = "fluid.change.pass";
		backgroundQuad = "true";
		shader = "fluid.change.shader";
		(textures) {
			pressurevelocity_old = "fluid.velocitypressure.texture";
			data = "fluid.data.texture";
		}
		(floats) {
			D_time = "D_time.float";
		}
		(vectors) {
			displacement = "displacement.vector";
			gravity = "gravity.vector";
		}
		(output) {
			rendertarget = "fluid.target";
			pressurevelocity_new = "fluid.copy.texture";
		}
	}

	(pass) {
		id = "fluid.change.copyback.pass";
		backgroundQuad = "true";
		shader = "copy.shader";
		(textures) {
			source = "fluid.copy.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			target = "fluid.velocitypressure.texture";
		}
	}

	//input

	(shader) {
		id = "input.shader";
		(vertex) {"
			#version 400 core

			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;
			out vec2 vposition;

			uniform vec4 position;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
				vposition = texcoords - position.xy;
			}
			"
		}
		(fragment) {"
			#version 400 core

			in vec2 vtexcoords;
			in vec2 vposition;
			out vec4 velocitypressure_new;

			uniform sampler2D velocitypressure_old;
			uniform vec4 velocity;

			void main() {
				vec4 pu = texture2D(velocitypressure_old,vtexcoords);
				if(vposition.x > -0.05 && vposition.y > -0.05 && vposition.x < 0.05 && vposition.y < 0.05) {
					velocitypressure_new = vec4(velocity.xy,0.0,pu.a);
				} else {
					velocitypressure_new = pu;
				}
			}
			"
		}
	}

	(pass) {
		id = "input.pass";
		backgroundQuad = "true";
		shader = "input.shader";
		(vectors) {
			velocity = "control.velocity.vector";
			position = "control.position.vector";
		}
		(textures) {
			velocitypressure_old = "fluid.velocitypressure.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			velocitypressure_new = "fluid.copy.texture";
		}
	}

	(pass) {
		id = "input.copyback.pass";
		backgroundQuad = "true";
		shader = "copy.shader";
		(textures) {
			source = "fluid.copy.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			target = "fluid.velocitypressure.texture";
		}
	}

	(effect) {
		id = "input.effect";
		(passes) {
			"input.pass"
			"input.copyback.pass"
		}
	}

	//advect

	(shader) {
		id = "advect.shader";
		(vertex) {"
			#version 400 core

			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;
			
			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
			}
			"
		}
		(fragment) {"
			#version 400 core

			in vec2 vtexcoords;
			out vec4 data_new;
			out vec4 pressurevelocity_new;

			uniform float D_time;
			uniform sampler2D data_old;
			uniform sampler2D pressurevelocity_old;
			
			void main() {

				//current values
				vec4 pu_l = texture2D(pressurevelocity_old,vtexcoords);
				float p_l = pu_l.a;
				vec2 u_l = pu_l.xy;
				vec4 data_l = texture2D(data_old,vtexcoords);

				//predictor step
				vec2 pos_p = vtexcoords - u_l * D_time;
				vec4 pu_p = texture2D(pressurevelocity_old,pos_p);
				float p_p = pu_p.a;
				vec2 u_p = pu_p.xy;
				vec4 data_p = texture2D(data_old,pos_p);
				
				//prepare corrector step
				vec2 pos_c = pos_p + u_p * D_time;
				vec4 pu_c = texture2D(pressurevelocity_old,pos_c);
				float p_c = pu_c.a;
				vec2 u_c = pu_c.xy;
				vec4 data_c = texture2D(data_old,pos_c);

				/*//corrector step
				pressurevelocity_new = clamp(vec4(u_p,0.0,p_p) + 0.5 * (vec4(u_l,0.0,p_l) - vec4(u_c,0.0,p_c)) , -10.0 , 10.0);
				data_new = clamp(data_p + 0.5 * (data_l - data_c) , 0.0 , 1.0);*/

				//calculate ranges
				float p_min = min( p_l , min( p_p , p_c ) );
				float p_max = max( p_l , max( p_p , p_c ) );
				vec2 u_min = min( u_l , min( u_p , u_c ) );
				vec2 u_max = max( u_l , max( u_p , u_c ) );
				vec4 data_min = min( data_l , min( data_p , data_c ) );
				vec4 data_max = max( data_l , max( data_p , data_c ) );

				//corrector step
				pressurevelocity_new = vec4( clamp( u_p + 0.5 * (u_l - u_c) , u_min , u_max ) , 0.0 , clamp( p_p + 0.5 * (p_l - p_c) , p_min , p_max ));
				data_new = clamp(data_p + 0.5 * (data_l - data_c) , data_min , data_max );
			}
			"
		}
	}

	(pass) {
		id = "advect.pass";
		backgroundQuad = "true";
		shader = "advect.shader";
		(floats) {
			D_time = "D_time.float";
		}
		(textures) {
			data_old = "fluid.data.texture";
			pressurevelocity_old = "fluid.velocitypressure.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			data_new = "fluid.copy.texture";
			pressurevelocity_new = "fluid.velocitypressure2.texture";
		}
	}

	(pass) {
		id = "fluid.data.copyback.pass";
		backgroundQuad = "true";
		shader = "copy.shader";
		(textures) {
			source = "fluid.copy.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			target = "fluid.data.texture";
		}
	}

	//poisson

	(shader) {
		id = "poisson.shader";
		(vertex) {"
			#version 400 core

			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
			}
			"
		}
		(fragment) {"
			#version 400 core

			in vec2 vtexcoords;
			out vec4 pressurevelocity_new;

			uniform sampler2D pressurevelocity_old;
			uniform sampler2D solid;
			uniform vec4 D_kernel;

			void main() {
				float s = abs(sign(texture2D(solid,vtexcoords).r));

				vec4 pl = texture2D(pressurevelocity_old,vtexcoords);
				float p = pl.a;
				vec2 u = pl.xy;

				if(s > 0.0) {
					pressurevelocity_new = vec4(vec3(0.0),p);
					return;
				}

				vec2 D_x = vec2(D_kernel.x,0);
				vec2 D_y = vec2(0,D_kernel.y);

				float s_xp = 1.0 - abs(sign(texture2D(solid,vtexcoords + D_x).r));
				float s_xm = 1.0 - abs(sign(texture2D(solid,vtexcoords - D_x).r));
				float s_yp = 1.0 - abs(sign(texture2D(solid,vtexcoords + D_y).r));
				float s_ym = 1.0 - abs(sign(texture2D(solid,vtexcoords - D_y).r));

				float p_xp = texture2D(pressurevelocity_old,vtexcoords + D_x).a;
				float p_xm = texture2D(pressurevelocity_old,vtexcoords - D_x).a;
				float p_yp = texture2D(pressurevelocity_old,vtexcoords + D_y).a;
				float p_ym = texture2D(pressurevelocity_old,vtexcoords - D_y).a;
				if(s_xp == 0.0) {
					p_xp = p;
				}
				if(s_xm == 0.0) {
					p_xm = p;
				}
				if(s_yp == 0.0) {
					p_yp = p;
				}
				if(s_ym == 0.0) {
					p_ym = p;
				}
				
				float u_xp = texture2D(pressurevelocity_old,vtexcoords + D_x).x * s_xp;
				float u_xm = texture2D(pressurevelocity_old,vtexcoords - D_x).x * s_xm;
				float u_yp = texture2D(pressurevelocity_old,vtexcoords + D_y).y * s_yp;
				float u_ym = texture2D(pressurevelocity_old,vtexcoords - D_y).y * s_ym;

				float p_new = ((p_xp + p_xm + p_yp + p_ym) - (u_xp - u_xm + u_yp - u_ym)) * 0.25;
				pressurevelocity_new = vec4(u,0,p_new);
			}
			"
		}
	}

	(pass) {
		id = "poisson.pass";
		backgroundQuad = "true";
		shader = "poisson.shader";
		(vectors) {
			D_kernel = "D_kernel.vector";
		}
		(textures) {
			pressurevelocity_old = "fluid.velocitypressure2.texture";
			solid = "solid.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			pressurevelocity_new = "fluid.velocitypressure3.texture";
		}
	}

	//project

	(shader) {
		id = "project.shader";
		(vertex) {"
			#version 400 core

			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
			}
			"
		}
		(fragment) {"
			#version 400 core

			in vec2 vtexcoords;
			out vec4 pressurevelocity_new;

			uniform sampler2D pressurevelocity_old;
			uniform sampler2D solid;
			uniform vec4 D_kernel;

			void main() {
				float s = abs(sign(texture2D(solid,vtexcoords).r));
				vec4 pu = texture2D(pressurevelocity_old,vtexcoords);
				vec2 u = pu.xy;
				float p = pu.a;
				if(s > 0.0) {
					pressurevelocity_new = vec4(vec3(0.0),p);
					return;
				}

				vec2 D_x = vec2(D_kernel.x,0);
				vec2 D_y = vec2(0,D_kernel.y);

				float s_xp = 1.0 - abs(sign(texture2D(solid,vtexcoords + D_x).r));
				float s_xm = 1.0 - abs(sign(texture2D(solid,vtexcoords - D_x).r));
				float s_yp = 1.0 - abs(sign(texture2D(solid,vtexcoords + D_y).r));
				float s_ym = 1.0 - abs(sign(texture2D(solid,vtexcoords - D_y).r));

				float p_xp = texture2D(pressurevelocity_old,vtexcoords + D_x).a;
				float p_xm = texture2D(pressurevelocity_old,vtexcoords - D_x).a;
				float p_yp = texture2D(pressurevelocity_old,vtexcoords + D_y).a;
				float p_ym = texture2D(pressurevelocity_old,vtexcoords - D_y).a;

				if(s_xm == 0.0) {
					p_xm = p;
				}
				if(s_yp == 0.0) {
					p_yp = p;
				}
				if(s_ym == 0.0) {
					p_ym = p;
				}

				vec2 gradient = vec2(
					p_xp - p_xm,
					p_yp - p_ym
					);

				vec2 u_new = u - gradient;
				u_new = vec2(u_new.x * s_xp * s_xm , u_new.y * s_yp * s_ym);
				pressurevelocity_new = vec4(u_new,0,p);
			}
			"
		}
	}

	(pass) {
		id = "project.pass";
		backgroundQuad = "true";
		shader = "project.shader";
		(vectors) {
			D_kernel = "D_kernel.vector";
		}
		(textures) {
			pressurevelocity_old = "fluid.velocitypressure3.texture";
			solid = "solid.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			pressurevelocity_new = "fluid.velocitypressure.texture";
		}
	}

	//draw something

	(vector) {
		id = "draw.origin.vector";
		(value) {"0 0 0"}
	}

	(shader) {
		id = "draw.shader";
		(vertex) {"
			#version 400 core
			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
			}
			"
		}
		(fragment) {"
			#version 400 core
			in vec2 vtexcoords;
			out vec4 paint;	

			uniform sampler2D original;
			uniform sampler2D image;

			uniform vec4 drawpos;

			void main() {
				vec4 orig = texture(original,vtexcoords);
				vec4 im = texture(image,vtexcoords + drawpos.xy);
				paint = orig*0.98 + im*0.02;
			}
			"
		}
	}
			
	(pass) {
		id = "draw.pass";
		backgroundQuad = "true";
		shader = "draw.shader";
		(vectors) {
			drawpos = "draw.origin.vector";
		}
		(textures) {
			original = "fluid.data.texture";
			image = "initfluid.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			paint = "fluid.copy.texture";
		}
	}

	(pass) {
		id = "draw.copyback.pass";
		backgroundQuad = "true";
		shader = "copy.shader";
		(textures) {
			source = "fluid.copy.texture";
		}
		(output) {
			rendertarget = "fluid.target";
			target = "fluid.data.texture";
		}
	}

	(effect) {
		id = "draw.effect";
		(passes) {
			"draw.pass"
			"draw.copyback.pass"
		}
	}

	//present and calculate

	(shader) {
		id = "present.shader";
		(vertex) {"
			#version 400 core

			in vec3 vertices;
			in vec2 texcoords;
			out vec2 vtexcoords;

			void main() {
				gl_Position = vec4(vertices,1);
				vtexcoords = texcoords;
			}
			"
		}
		(fragment) {"
			#version 400 core
			
			in vec2 vtexcoords;
			out vec4 fragment;

			uniform sampler2D data;
			uniform sampler2D solid;
			uniform sampler2D pressurevelocity;
			uniform sampler2D shadowmap;
			uniform float present;

			void main() {
				if(present == 1.0) {
					vec4 s = texture2D(solid,vtexcoords);
					s = clamp(vec4((s.r + s.g + s.b)*10.0) , 0 , 1);
					vec4 background = max(s , texture2D(data,vtexcoords));
					fragment = background * (1.0 - texture2D(shadowmap,clamp(vtexcoords + vec2(-0.005,0.005),0,1)).r*0.3);;
				} else if(present == 0.0) {
					//fragment = abs(texture2D(pressurevelocity,vtexcoords));
//					fragment.r = abs(texture2D(pressurevelocity,vtexcoords).a);

					//show vorticity map
					vec2 dx = vec2(0.01,0);
					vec2 dy = vec2(0,0.01);
					float vorticity = 
						texture2D(pressurevelocity,vtexcoords + dx).y - texture2D(pressurevelocity,vtexcoords - dx).y
						- (
						texture2D(pressurevelocity,vtexcoords + dy).x - texture2D(pressurevelocity,vtexcoords - dy).x
						);
					vorticity = clamp(abs(vorticity) * 100.0 , 0.0 , 1.0);
					
					vec4 vort = vorticity * vec4(1,0.7,0.7,1) + (1.0-vorticity) * vec4(0.7,1,0.7,1);
					vec4 s = texture2D(solid,vtexcoords);
					s = clamp(vec4((s.r + s.g + s.b)*10.0) , 0 , 1);
					vec4 background = max(s , vort);
					
					fragment = background * (1.0 - texture2D(shadowmap,clamp(vtexcoords + vec2(-0.005,0.005),0,1)).r*0.3);
				}
			}
			"
		}
	}

	(rendertarget) {
		id = "present.offline.target";
		width = 300;
		height = 200;
	}

	(texture) {
		id = "glyphshadow.texture";
		width = 300;
		height = 200;
	}

	(texture) {
		id = "glyphshadow2.texture";
		width = 300;
		height = 200;
	}

	(rendertarget) {
		id = "present.rendertarget";
		width = 300;
		height = 200;
		renderToDisplay = "true";
	}

	(pass) {
		id = "present.pass";
		shader = "present.shader";
		backgroundQuad = "true";
		(floats) {
			present = "present.float";
		}
		(textures) {
			data = "fluid.data.texture";
			pressurevelocity = "fluid.velocitypressure.texture";
			solid = "solid.texture";
			shadowmap = "glyphshadow.texture";
		}
		(output) {
			rendertarget = "present.rendertarget";
		}
	}

	(pass) {
		id = "presentarrow.pass";
		wireframe="false";
		(output) {
			rendertarget = "present.rendertarget";
			clearDepthBuffer = "false";
			clearColorBuffer = "false";
			targetBlendFactor = "one_minus_fragmentalpha";
			fragmentBlendFactor = "fragmentalpha";
		}
	}

	(effect) {
		id = "calculate.effect";
		(passes) {
			"fluid.change.pass"
			"fluid.change.copyback.pass"
			"advect.pass"
			"fluid.data.copyback.pass"
			"poisson.pass"
			"project.pass"

			"pos.advect.pass"
			"pos.copyback.pass"
			"vel.pass"
		}
	}

	(effect) {
		id = "present.effect";
		(passes) {
			"glyphshadow.pass"
			"gauss.x.pass"
			"gauss.y.pass"
			"present.pass"
			"presentarrow.pass"
		}
	}

}
